"""Module containing for SOI-module 'Autentifiseringstavle'."""
import string
from random import choices
from PySide2.QtWidgets import QTableWidget, QTableWidgetItem
from PySide2 import QtGui
from PySide2.QtCore import Qt
from soitool.modules.module_base import (
    ModuleBase,
    get_table_size,
    resize_table,
    HEADLINE_FONT,
)

START_NO_OF_AUTHENTICATION_CODES = 10
CODE_LENGTH = 26
CODE_CHARACTERS = string.ascii_uppercase  # Codes consist of these characters
ROW_IDENTIFIERS = string.ascii_uppercase  # Characters for second column,
# it's length determines maximum number of codes (rows).
SEPARATE_INTERVAL = 5  # Adds space between sets of characters, 0 => no spaces
# If code is 123456 and interval is 2, code will be 12 34 56
# If code is 1234567 and interval is 2, code will be 123 45 67

HEADLINE_TEXT = "Autentifiseringstavle"


class Meta(type(ModuleBase), type(QTableWidget)):
    """Used as a metaclass to enable multiple inheritance."""


class AuthentificationBoardModule(ModuleBase, QTableWidget, metaclass=Meta):
    """Modified QTableWidget representing a 'Autentifiseringstavle'.

    By default, the widget initializes with a headline, a row-count of
    START_NO_OF_AUTHENTIFICATION_CODES and three columns.
    Row x in the first column contains the character CODE_CHARACTERS[x].
    Row x in the second column contains the character x.
    Row x in the third column contains an authentification code.

    If parameters are given, the widget initializes accordingly:
    'size' is a dict: {"width": int, "height": int},
    'data' is a 2D list where data[x][y] represents row x, column y, and
    data[0][0] is the headline.

    The widget does not use more room than needed, and resizes dynamically.
    It has shortcuts for adding and removing rows.

    Inherits from ModuleBase and QTableWidget.
    ModuleBase is used as an interface, it's methods are overridden.
    """

    def __init__(self, size=None, data=None):
        self.type = "AuthentificationBoardModule"
        QTableWidget.__init__(self)
        ModuleBase.__init__(self)

        # Remove headers and scrollbars
        self.horizontalHeader().hide()
        self.verticalHeader().hide()
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        # If parameters are None, generate new table
        if size is None and data is None:
            self.generate_table_data()
            self.resizeColumnsToContents()

            self.insert_headline()
            # Resize height of rows and set size of window
            resize_table(self, resize_column=False)
        else:
            self.setColumnCount(len(data[1]))
            self.setRowCount(len(data) - 1)  # - 1 to skip headline

            # Set cell-items
            for i in range(self.rowCount()):
                for j in range(self.columnCount()):
                    item = QTableWidgetItem(data[i + 1][j])  # +1 skip headline
                    self.setItem(i, j, item)

            self.resizeColumnsToContents()
            resize_table(self, resize_column=False)
            self.setFixedWidth(size["width"])
            self.setFixedHeight(size["height"])

            self.insert_headline(data[0])

    def generate_table_data(self):
        """Insert authentification codes and row identifiers."""
        # Set number of rows and columns
        self.setRowCount(START_NO_OF_AUTHENTICATION_CODES)
        self.setColumnCount(3)

        # Get generated codes
        codes = generate_authentification_codes()

        # Insert table data
        for i in range(START_NO_OF_AUTHENTICATION_CODES):
            # Insert row identifier in first column
            item_first = QTableWidgetItem(ROW_IDENTIFIERS[i])
            item_first.setFlags(item_first.flags() ^ Qt.ItemIsEditable)
            self.setItem(i, 0, item_first)

            # Insert row identifier (int) in second column
            item_second = QTableWidgetItem(str(i))
            item_second.setFlags(item_second.flags() ^ Qt.ItemIsEditable)
            self.setItem(i, 1, item_second)

            # Insert code in third column
            item_third = QTableWidgetItem(codes[i])
            item_third.setFlags(item_third.flags() ^ Qt.ItemIsEditable)
            self.setItem(i, 2, item_third)

    def insert_headline(self, text=HEADLINE_TEXT):
        """Insert headline text.

        Parameters
        ----------
        text : string, optional
            The headline text, by default HEADLINE_TEXT
        """
        item_headline = QTableWidgetItem(text)
        item_headline.setTextAlignment(Qt.AlignHCenter)
        item_headline.setFont(HEADLINE_FONT)

        self.insertRow(0)
        self.setItem(0, 0, item_headline)
        self.setSpan(0, 0, 1, self.columnCount())  # Make item span all columns

    def generate_unique_authentification_code(self):
        """Generate a authentification-code that does not already exist."""
        # Get existing codes
        existing_codes = self.get_codes()

        # Randomly generate a new code until it is unique
        unique_code = False
        while not unique_code:
            code = choices(CODE_CHARACTERS, k=CODE_LENGTH)
            unique_code = code not in existing_codes

        # Choose starting point for separating code with spaces, depending on
        # SEPARATE_INTERVAL and CODE_LENGTH being even or odd numbers.
        # If code is 1234567 and interval is 3, code will be 1234 567 instead
        # of 123 456 7 after space separation.
        if (
            SEPARATE_INTERVAL % 2 == 0
            and CODE_LENGTH % 2 == 0
            or SEPARATE_INTERVAL % 2 != 0
            and CODE_LENGTH % 2 != 0
        ):
            start = SEPARATE_INTERVAL - 1
        else:
            start = SEPARATE_INTERVAL

        # Separate code with spaces
        for i in range(start, len(code), SEPARATE_INTERVAL):
            code[i] += " "

        return "".join(code)

    def keyPressEvent(self, event):
        """Add or remove row when 'Ctrl + +' and 'Ctrl + -' are pressed."""
        if (
            event.modifiers() == Qt.ControlModifier
            and event.key() == Qt.Key_Plus
        ):
            self.add_row()
        elif (
            event.modifiers() == Qt.ControlModifier
            and event.key() == Qt.Key_Underscore
        ):
            self.remove_row()
        else:
            super(AuthentificationBoardModule, self).keyPressEvent(event)

    def add_row(self):
        """Add row with cell-data below selected row."""
        row_index = self.currentRow()
        # If maximum amount of rows not reached and a row is selected
        # (+ 1 to skip row containing headline)
        if self.rowCount() < len(ROW_IDENTIFIERS) + 1 and row_index != -1:
            # Generate unique code and insert row
            code = self.generate_unique_authentification_code()
            self.insertRow(row_index + 1)

            # Loop through all rows starting with the new row
            for i in range(row_index + 1, self.rowCount()):
                # Insert row identifier in first column
                item_first = QTableWidgetItem(CODE_CHARACTERS[i - 1])
                item_first.setFlags(item_first.flags() ^ Qt.ItemIsEditable)
                self.setItem(i, 0, item_first)

                # Insert row identifier (int) in second column
                item_second = QTableWidgetItem(str(i - 1))
                item_second.setFlags(item_second.flags() ^ Qt.ItemIsEditable)
                self.setItem(i, 1, item_second)

            # Insert authentification-code in third column
            item_third = QTableWidgetItem(code)
            item_third.setFlags(item_third.flags() ^ Qt.ItemIsEditable)
            self.setItem(row_index + 1, 2, item_third)

            # Resize code-column in case it got wider
            # Example: Letters BGD is wider than I (depending on font)
            self.resizeColumnToContents(2)
            # Resize widget size
            resize_table(self, resize_column=False)

    def remove_row(self):
        """Remove selected row."""
        row_index = self.currentRow()
        # If at least one row exists (+ headline) and a row is selected
        if self.rowCount() > 2 and row_index != 0 and row_index != -1:
            # Remove row
            self.removeRow(row_index)
            # 'Decrease' row identifiers below the removed row
            # If first row is removed, identifier A,B,C becomes A,B (not B,C)
            for i in range(row_index, self.rowCount()):
                self.item(i, 0).setText(CODE_CHARACTERS[i - 1])
                self.item(i, 1).setText(str(i - 1))
        # Resize widget size
        resize_table(self, resize_column=False)

    def get_codes(self):
        """Get all authentification-codes in table.

        Returns
        -------
        List
            List containing authentification-codes.
        """
        codes = []

        # Start with row 1 to skip headline
        for i in range(1, self.rowCount()):
            codes.append(self.item(i, 2).text())

        return codes

    def get_size(self):
        """Get size of widget."""
        return get_table_size(self)

    def get_data(self):
        """Return list containing all data.

        Returns
        -------
        List (2D)
            list[0][0] contains headline,
            list[x][y] represents value of row x, column y.
        """
        content = []
        item_headline = self.item(0, 0)
        if item_headline is not None:
            content.append(item_headline.text())

        for i in range(1, self.rowCount()):
            row = []
            for j in range(self.columnCount()):
                row.append(self.item(i, j).text())
            content.append(row)

        return content

    @staticmethod
    def get_user_friendly_name():
        """Get user-friendly name of module."""
        return "Autentifiseringstavle"

    @staticmethod
    def get_icon():
        """Get icon of module."""
        return QtGui.QIcon("soitool/media/authentificationboardmodule.png")


def generate_authentification_codes(amount=START_NO_OF_AUTHENTICATION_CODES):
    """Generate and return a specified amount of authentification-codes.

    Parameters
    ----------
    amount : int, optional
        Number of codes to generate,
        by default START_NO_OF_AUTHENTICATION_CODES

    Returns
    -------
    List
        List containing generated authentification-codes.
    """
    codes = []
    for _ in range(amount):
        sequence = choices(CODE_CHARACTERS, k=CODE_LENGTH)
        # Choose starting point for separating code with spaces, depending on
        # SEPARATE_INTERVAL and CODE_LENGTH being even or odd numbers.
        # If code is 1234567 and interval is 3, code will be 1234 567 instead
        # of 123 456 7 after space separation.
        if (
            SEPARATE_INTERVAL % 2 == 0
            and CODE_LENGTH % 2 == 0
            or SEPARATE_INTERVAL % 2 != 0
            and CODE_LENGTH % 2 != 0
        ):
            start = SEPARATE_INTERVAL - 1
        else:
            start = SEPARATE_INTERVAL
        for j in range(
            start, len(sequence) - SEPARATE_INTERVAL, SEPARATE_INTERVAL
        ):
            sequence[j] += " "
        codes.append("".join(sequence))

    return codes
